package hr.fer.bachelorThesis.algorithm.pso;

import java.util.List;

import hr.fer.bachelorThesis.algorithm.memory.ISharedMemory;
import hr.fer.bachelorThesis.labyrinth.ILabyrinth;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.Vector;

/**
 * Evolutionary algorithm agent structure. It serves as a container for all the
 * relevant information for the algorithm to work and succeed in optimisation.
 * 
 * @author Viran
 *
 */
public class Agent implements IAgent {

	/**
	 * Best agent in population by its fitness.
	 */
	private IAgent popBestAgent;

	/**
	 * Goal position this agent is trying to reach.
	 */
	private IVector targetPos;

	/**
	 * Current agent position.
	 */
	private IVector curPos;

	/**
	 * Previous agent position.
	 */
	private IVector prevPos;

	/**
	 * Best previous agent position.
	 */
	private IVector persBestPrevPos;

	/**
	 * Best positions fitness.
	 */
	private double bestPrevFitness;

	/**
	 * The direction in which agent is currently moving.
	 */
	private IVector curDirVector;

	/**
	 * Fitness of this agent.
	 */
	double fitness;

	/**
	 * Number of steps this agent has taken while searching for the exit.
	 */
	private double totalPositionsPassed;

	/**
	 * Represents the quality of the positions found so far.
	 */
	private double qualityOfPositionsFound;

	/**
	 * Flag for checking weather this agent has finished searching.
	 */
	private boolean isFinished;

	/**
	 * Unique identification number.
	 */
	private long ID;

	private boolean colidedRecently;

	/**
	 * Agent constructor.
	 * 
	 * @param uniqueAgentId
	 *            Agent id number.
	 * @param startingPosition
	 *            Agent starting position.
	 * @param directionVector
	 *            Agent starting direction vector.
	 */
	public Agent(long uniqueAgentId, IVector startingPosition, IVector directionVector) {
		this.ID = uniqueAgentId;
		this.curPos = startingPosition;
		this.curDirVector = directionVector;
		this.totalPositionsPassed = 1;
		this.qualityOfPositionsFound = 0;
	}

	@Override
	public int compareTo(IAgent o) {
		return Double.compare(this.fitness, o.getFitness());
	}

	@Override
	public void setPopBestAgent(IAgent popBestAgent) {
		this.popBestAgent = popBestAgent;
	}

	@Override
	public IAgent getPopBestAgent() {
		return this.popBestAgent;
	}

	@Override
	public IVector getPersonalTargetPosition() {
		return this.targetPos;
	}

	@Override
	public void setPersonalTargetPosition(IVector targetPosition) {
		this.targetPos = targetPosition;
	}

	@Override
	public IVector getCurrentPosition() {
		return this.curPos;
	}

	@Override
	public IVector getPreviousPosition() {
		return this.prevPos;
	}

	@Override
	public IVector getBestPreviousPosition() {
		return this.persBestPrevPos;
	}

	@Override
	public double getPrevBestFitness() {
		return this.bestPrevFitness;
	}

	@Override
	public IVector getCurrentDirectionVector() {
		return this.curDirVector;
	}

	@Override
	public double getFitness() {
		return this.fitness;
	}

	@Override
	public boolean getIsFinished() {

		return this.isFinished;
	}

	@Override
	public boolean colided() {
		return this.colidedRecently;
	}
	
	@Override
	public long getID() {
		return this.ID;
	}

	@Override
	public IVector move(IVector directionVector, ILabyrinth labyrinth, ISharedMemory popMem) {
		double beforeMovMagnitude=directionVector.norm();
		List<IVector> positions = labyrinth.generatePosition(this.curPos, directionVector);
		double afterMovMagnitude=directionVector.norm();
		
		if(beforeMovMagnitude!=afterMovMagnitude)
			this.colidedRecently=true;
		else
			this.colidedRecently=false;
		
		if (!positions.isEmpty())
			positions.remove(0);
		for (IVector position : positions) {
			this.qualityOfPositionsFound += popMem.steppedOn(position);
			this.totalPositionsPassed++;
		}
		// remember n-1 position
		this.prevPos = curPos;
		// process for this iteration: new position, and fitness
		if (!positions.isEmpty())
			this.curPos = positions.get(positions.size() - 1); // else stay at
																// same position
		double nextPosFitness = qualityOfPositionsFound / totalPositionsPassed;
		this.fitness = nextPosFitness;
		// previous best position
		if (this.bestPrevFitness <= this.fitness) {
			this.bestPrevFitness = fitness;
			this.persBestPrevPos = curPos;
		}
		// is finished
		this.isFinished = labyrinth.isFinished(curPos);
		/// direction vector

		this.curDirVector = directionVector;

		return this.curPos;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (int) (ID ^ (ID >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Agent other = (Agent) obj;
		if (ID != other.ID)
			return false;
		return true;
	}

	

}
