package hr.fer.bachelorThesis.algorithm.pso;

import hr.fer.bachelorThesis.algorithm.memory.ISharedMemory;
import hr.fer.bachelorThesis.labyrinth.ILabyrinth;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.Vector;

/**
 * Interface for a evolutionary algorithm agent.
 * @author Viran
 *
 * @param <IVector> Point of domain in which this agent operates. 
 */
public interface IAgent extends Comparable<IAgent>{
	
	/**
	 * Set the best agent in population.
	 * @param popBestAgent Best agent in population.
	 */
	public void setPopBestAgent(IAgent popBestAgent);
	
	/**
	 * Get best agent in population.
	 * @return Best agent in population.
	 */
	public IAgent getPopBestAgent();
	
	/**
	 * Get personal target position.
	 * @return Personal target position.
	 */
	public IVector getPersonalTargetPosition();
	
	/**
	 * Set personal target position.
	 * @param goalPosition Personal target position.
	 */
	public void setPersonalTargetPosition(IVector targetPosition);
	
	/**
	 * Get current position.
	 * @return Current position.
	 */
	public IVector getCurrentPosition();
	
	/**
	 * Get previous position.
	 * @return PRevious position.
	 */
	public IVector getPreviousPosition();
	
	/**
	 * Get Best previous position.
	 * @return Best previous position.
	 */
	public IVector getBestPreviousPosition();
	
	/**
	 * Get agent fitness value at the best previous position.
	 * @return Fitness value.
	 */
	public double getPrevBestFitness();
	
	/**
	 * Get current direction vector.
	 * @return Current direction vector.
	 */
	public IVector getCurrentDirectionVector();
	
	/**
	 * Get agent fitness value.
	 * @return Fitness value.
	 */
	public double getFitness();
	
	/**
	 * Get is the agent has found the exit or is done with iterations.
	 * @return True if the agent is done searching or false otherwise.
	 */
	public boolean getIsFinished();
	
	/**
	 * Get this agent's unique identifier in population. TODO make it in algorithm!
	 * @return A unique agent identifier.
	 */
	public long getID();

	/**
	 * For the given direction vector, move the agent across the labyrinth while updating the population memory.
	 * @param directionVector Vector of the direction this agent should move.
	 * @param labyrinth Labyrinth in which the agent searches for the exit.
	 * @param popMem Population memory which he uses while navigating.
	 * @return The next position at which the agent has arrived.
	 */
	public IVector move(IVector directionVector, ILabyrinth labyrinth, ISharedMemory popMem);

	public boolean colided();
	
}
