package hr.fer.bachelorThesis.algorithm.pso;

import java.util.LinkedList;
import java.util.concurrent.BlockingQueue;

import hr.fer.bachelorThesis.algorithm.IAlgorithm;
import hr.fer.bachelorThesis.algorithm.IPopulation;
import hr.fer.bachelorThesis.algorithm.memory.ITargetSystemMemory;
import hr.fer.bachelorThesis.algorithm.stopCondition.IStopCondition;
import hr.fer.bachelorThesis.algorithm.strategy.IAgentStrategy;
import hr.fer.bachelorThesis.controller.AgentIterationLog;
import hr.fer.bachelorThesis.controller.MovementSplitter;
import hr.fer.bachelorThesis.controller.ThreadMonitor;
import hr.fer.bachelorThesis.controller.ThreadMonitor.Label;
import hr.fer.bachelorThesis.labyrinth.ILabyrinth;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.IncompatibleOperandException;

public class PSOAlgorithm implements IAlgorithm{

	protected long id;
	protected IStopCondition stopCondition;
	protected ILabyrinth labyrinth;
	protected IPopulation population;
	protected ITargetSystemMemory populationMemory;
	protected IAgentStrategy agentStrategy;
	protected boolean setToDraw;
	protected BlockingQueue<PSOIterationData> algorithmControllerQ;
	protected ThreadMonitor monitor;
	protected MovementSplitter splitter;
	private LinkedList<Long> undrawableAgnet=new LinkedList<>();
	
	public PSOAlgorithm(long id,ILabyrinth labyrinth, IAgentStrategy agentStrategy, IPopulation population,
			IStopCondition stopCondition) {
		this.id=id;
		this.stopCondition = stopCondition;
		this.labyrinth = labyrinth;
		this.population = population;
		this.agentStrategy = agentStrategy;
		this.setToDraw=false;
		this.algorithmControllerQ=null;
		this.splitter=new MovementSplitter();
	}



	@Override
	public void run() {
		this.stopCondition.initialise();
		while (!stopCondition.isDone()) {
			this.algorithmIteration();
			this.stopCondition.iterationPassed();
			if(setToDraw)
				this.sendData();
			//Population refresh must go after population
			this.population.refreshPopulationState();
			if(this.checkMonitor())
				break;
		}
	}

	private boolean checkMonitor() {
		synchronized (monitor) {
			if (monitor.getLabel() == Label.PAUSE) {

				while (monitor.getLabel() == Label.PAUSE) {
					try {
						monitor.wait();
					} catch (InterruptedException e) {
					}
				}

				if (monitor.getLabel() == Label.STOP) {
					
					return true;
				} else if (monitor.getLabel() == Label.RUN) {
					
					return false;
				}
				
			} else if (monitor.getLabel() == Label.STOP) {
				return true;
			}
		}
		return false;
	}

	private void algorithmIteration() {
		IVector dirVector = null;
		for (IAgent agent : population.unfinishedAgentsIterable()) {
			try {
				dirVector = this.agentStrategy.nextDirectionVector(agent);
			} catch (IncompatibleOperandException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			agent.move(dirVector, labyrinth, populationMemory);
			//System.out.println("Agent "+agent.getID()+" @ "+agent.getCurrentPosition()+" -> "+agent.getCurrentDirectionVector().norm());
		}
		
		//System.out.println(populationMemory);
	}

	private void sendData() {
		//Split positions and send them to the drawer.
		PSOIterationData newLog=new PSOIterationData();
		
		newLog.algorithmId=this.id;
		newLog.algorithmIteration=stopCondition.getCurrentIteration();
		newLog.bestAgentId=population.globalBestAgent().getID();
		double maxSpeed=1;
		for(IAgent agent:population.allAgentsIterable()){
			if(!this.undrawableAgnet.contains(agent.getID())){
				newLog.agentsLog.add(new AgentIterationLog(agent.getID(),agent.getCurrentPosition().copy(),agent.getPreviousPosition().copy(),agent.getCurrentDirectionVector().copy()));
				if(maxSpeed<agent.getCurrentDirectionVector().norm()) maxSpeed=agent.getCurrentDirectionVector().norm();
				if(agent.getIsFinished())
					this.undrawableAgnet.add(agent.getID());
			}
		}
		newLog.maxSpeed=maxSpeed;
		try {
			this.algorithmControllerQ.put(newLog);
		} catch (InterruptedException e) {
			
			return;
		}
	}
	
	@Override
	public void setCommunicationQueue(BlockingQueue<PSOIterationData> queue) {
		this.algorithmControllerQ=queue;
		this.setToDraw=true;
	}

	@Override
	public long getID() {
		return this.id;
	}
	
	@Override
	public void setID(long id) {
		this.id=id;
	}

	@Override
	public void setMonitor(ThreadMonitor monitor) {
		this.monitor=monitor;
	}

	@Override
	public ITargetSystemMemory getSharedMemory() {
		return this.populationMemory;
	}

	@Override
	public void setSharedMemory(ITargetSystemMemory populationMemory) {
		this.populationMemory = populationMemory;
	}



	

}
