package hr.fer.bachelorThesis.algorithm.pso;

import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import hr.fer.bachelorThesis.algorithm.IAlgorithm;
import hr.fer.bachelorThesis.algorithm.IAlgorithmDrawer;
import hr.fer.bachelorThesis.algorithm.IAlgorithmManager;
import hr.fer.bachelorThesis.algorithm.memory.ITargetSystemMemory;
import hr.fer.bachelorThesis.controller.AgentIterationLog;
import hr.fer.bachelorThesis.controller.AgentSubIterationLog;
import hr.fer.bachelorThesis.controller.MovementSplitter;
import hr.fer.bachelorThesis.controller.ThreadMonitor;
import hr.fer.bachelorThesis.controller.ThreadMonitor.Label;
import hr.fer.bachelorThesis.gui.labyrinth.LabyrinthView;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

public class PSOAlgorithmManager implements IAlgorithmManager {

	private static final int ALG_BUFFER_SIZE = 10;
	private static final int DRAW_BUFFER_SIZE = 40; //TODO change back to 20 maybe?

	private PSOAlgorithm psoAlgorithm;
	private PSODrawer psoDrawer;

	private BlockingQueue<PSOIterationData> agentControllerQ;
	private BlockingQueue<PSOSubIterationData> controllerDrawerQ;
	// private BlockingQueue<PSOIterationData> alg2DrawQ;

	private Thread algThread;
	private Thread drawThread;

	private ThreadMonitor algMonitor;
	private ThreadMonitor drawMonitor;

	private MovementSplitter splitter = new MovementSplitter();
	private boolean isDone;
	private Color agentPaint;
	private Color bestAgentPaint;
	private boolean displayMemory=false;

	public PSOAlgorithmManager(PSOAlgorithm algorithm,Color agentPaint,Color bestAgentPaint) {
		this.agentControllerQ = new LinkedBlockingQueue<>(ALG_BUFFER_SIZE);
		this.controllerDrawerQ = new LinkedBlockingQueue<>(DRAW_BUFFER_SIZE);
		// this.alg2DrawQ=new LinkedBlockingQueue<>(BUFFER_SIZE);

		this.algMonitor = new ThreadMonitor(Label.STOP);
		this.drawMonitor = new ThreadMonitor(Label.STOP);

		this.psoAlgorithm = algorithm;
		this.psoAlgorithm.setCommunicationQueue(this.agentControllerQ);
		// this.psoAlgorithm.setCommunicationQueue(this.alg2DrawQ);

		this.psoAlgorithm.setMonitor(algMonitor);
		this.isDone = false;
		
		this.agentPaint=agentPaint;
		this.bestAgentPaint=bestAgentPaint;
	}

	@Override
	public IAlgorithm getAlgorithm() {
		return this.psoAlgorithm;
	}

	@Override
	public IAlgorithmDrawer getAlgorithmDrawer() {
		return this.psoDrawer;
	}

	@Override
	public void startAlgorithm() {
		this.algThread = new Thread(psoAlgorithm);
		this.drawThread = new Thread(psoDrawer);

		this.algMonitor.setLabel(Label.RUN);
		this.drawMonitor.setLabel(Label.RUN);

		this.algThread.start();
		this.drawThread.start();
	}

	@Override
	public void pauseAlgorithm() {
		synchronized (algMonitor) {
			algMonitor.setLabel(Label.PAUSE);
		}
		synchronized (drawMonitor) {
			drawMonitor.setLabel(Label.PAUSE);
		}

	}

	@Override
	public void resumeAlgorithm() {
		synchronized (algMonitor) {
			algMonitor.setLabel(Label.RUN);
			algMonitor.notify();
		}
		synchronized (drawMonitor) {
			drawMonitor.setLabel(Label.RUN);
			drawMonitor.notify();
		}
	}

	@Override
	public void stopAlgorithm() {
		synchronized (algMonitor) {
			algMonitor.setLabel(Label.STOP);
			algMonitor.notify();
		}
		synchronized (drawMonitor) {
			drawMonitor.setLabel(Label.STOP);
			drawMonitor.notify();
		}

		algThread.interrupt();
		drawThread.interrupt();
		// this removes all threads from monitor waiting queues
		try {
			algThread.join();
			drawThread.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		this.isDone = true;
	}

	@Override
	public void setLabyrinthView(LabyrinthView labyrinthView, double agentDim) {
		// this.psoDrawer=new PSODrawer(this.alg2DrawQ,labyrinthView, agentDim);
		this.psoDrawer = new PSODrawer(this.psoAlgorithm.getID(),this.controllerDrawerQ, labyrinthView, agentDim, this.agentPaint,this.bestAgentPaint, this.displayMemory);
		this.psoDrawer.setMonitor(drawMonitor);
	}

	@Override
	public void setSharedMemory(IAlgorithmManager firstManager) {
		this.psoAlgorithm.setSharedMemory(firstManager.getSharedMemory());

	}

	@Override
	public ITargetSystemMemory getSharedMemory() {
		return this.psoAlgorithm.getSharedMemory();
	}

	@Override
	public boolean drawingBufferEmpty() {
		boolean res = false;
		synchronized (controllerDrawerQ) {
			res = controllerDrawerQ.isEmpty();
		}
		return res;
	}

	@Override
	public void prepareNextIteration() {
		if (algThread.isAlive()) {
			proccessNextIteration();
		} else {
			emptyAlgBuffer();
			emptyDrawBuffer();
			this.isDone = true;
			// this.stopAlgorithm();
		}

	}

	private void emptyDrawBuffer() {
		try {
			synchronized (drawMonitor) {
				drawMonitor.setLabel(Label.STOP);
				drawMonitor.notify();
			}
			this.drawThread.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void emptyAlgBuffer() {
		while (!this.agentControllerQ.isEmpty()) {
			LinkedList<IVector[]> agentSubPositions = new LinkedList<>();
			PSOIterationData iterData = this.agentControllerQ.poll();
			if (iterData == null)
				return;
			for (AgentIterationLog agentLog : iterData.agentsLog) {
				IVector[] agentsSubIter = splitter.splitMovement(agentLog.previousPosition, agentLog.currentPosition,
						(int) iterData.maxSpeed, psoDrawer.getAgentDim(), psoDrawer.getAgentDim());
				agentSubPositions.add(agentsSubIter);

			}

			processOutput(agentSubPositions, iterData);

		}
		try {
			this.algThread.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void proccessNextIteration() {
		LinkedList<IVector[]> agentSubPositions = new LinkedList<>();
		PSOIterationData iterData = null;
		try {
			iterData = this.agentControllerQ.take();
		} catch (InterruptedException e) {
			return;
		}
		if (iterData == null)
			return;

		for (AgentIterationLog agentLog : iterData.agentsLog) {
			IVector[] agentsSubIter = splitter.splitMovement(agentLog.previousPosition, agentLog.currentPosition,
					(int) iterData.maxSpeed, psoDrawer.getAgentDim(), psoDrawer.getAgentDim());
			agentSubPositions.add(agentsSubIter);

		}

		processOutput(agentSubPositions, iterData);
	}

	private void processOutput(LinkedList<IVector[]> agentSubPositions, PSOIterationData iterData) {
		int splitPos;
		try {
			splitPos = agentSubPositions.getFirst().length;
		} catch (NoSuchElementException | NullPointerException e) {
			return;
		}
		for (int i = 0; i < splitPos; i++) {
			PSOSubIterationData subIterData = new PSOSubIterationData();
			subIterData.algorithmId = iterData.algorithmId;
			subIterData.algorithmIteration = iterData.algorithmIteration;
			subIterData.bestAgentId = iterData.bestAgentId;
			int counter = 0;
			for (IVector[] vectors : agentSubPositions) {
				subIterData.agentsLog
						.add(new AgentSubIterationLog(iterData.agentsLog.get(counter).agentId, vectors[i]));
				counter++;
			}

			try {
				this.controllerDrawerQ.put(subIterData);
			} catch (InterruptedException e) {
				return;
			}

		}
	}

	@Override
	public boolean isDone() {
		return this.isDone;
	}

	@Override
	public void shouldDisplayMemory(boolean displayMemory) {
		this.displayMemory=displayMemory;
	}

}
