package hr.fer.bachelorThesis.algorithm.pso;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import hr.fer.bachelorThesis.algorithm.AgentIdGenerator;
import hr.fer.bachelorThesis.algorithm.IPopulation;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.IncompatibleOperandException;
import hr.fer.bachelorThesis.util.linearAlgebra.Vector;
import hr.fer.bachelorThesis.util.random.IRNG;
import hr.fer.bachelorThesis.util.random.RNG;

public class Population implements IPopulation {

	protected List<IAgent> agents;
	protected int globalBestIndex;
	protected IRNG rand;
	protected int halfNeighbourRadius;
	private double maxPopulationSpeed;
	private IVector popVector;
	private boolean oddNeighNum = false;
	private IVector centerPosition;
//	private double minDistFromCenver;
//	private double maxDistFromCenver;

	public Population(int numberOfAgents, IVector startingPositions, double maxInitSpeed, double maxPopulationSpeed,
			int neighboursPerAgent) {
		this.rand = RNG.getRNG();
		this.agents = new ArrayList<IAgent>();
		this.globalBestIndex = 0;
		this.maxPopulationSpeed = maxPopulationSpeed;
		this.halfNeighbourRadius = (neighboursPerAgent) / 2;
		if (neighboursPerAgent % 2 == 1)
			this.oddNeighNum = true;
		this.popVector = new Vector(0, 0);
		for (int i = 0; i < numberOfAgents; i++) {
			double randomAngle = (2 * Math.PI) * rand.nextDouble();
			double randomAmp = maxInitSpeed * rand.nextDouble();
			IVector dirVector = new Vector(randomAmp * Math.cos(randomAngle), randomAmp * Math.sin(randomAngle));
			agents.add(i, new Agent(AgentIdGenerator.getUniqueAgentId(), startingPositions, dirVector));
		}
		centerPosition = new Vector(0, 0);
		refreshPopulationVector();
		refreshPopulationCenter();
	}

	public Collection<IAgent> getAgentNeighbours(IAgent agent) {
		Set<IAgent> neighbours = new HashSet<>();
		int givenAgentIndex = agents.indexOf(agent);
		int nextInList;
		int neighIndex;
		for (neighIndex = -this.halfNeighbourRadius; neighIndex <= this.halfNeighbourRadius; neighIndex++) {
			if (neighIndex == 0)
				continue;
			nextInList = givenAgentIndex + neighIndex;
			if (nextInList < 0)
				nextInList = agents.size() + nextInList;
			if (nextInList > agents.size() - 1)
				nextInList = nextInList % agents.size();
			neighbours.add(agents.get(nextInList));
		}

		if (this.oddNeighNum) {
			if (neighIndex != 0) {
				nextInList = givenAgentIndex + neighIndex;
				if (nextInList < 0)
					nextInList = agents.size() + nextInList;
				if (nextInList > agents.size() - 1)
					nextInList = nextInList % agents.size();
				neighbours.add(agents.get(nextInList));
			}
		}

		return neighbours;
	}

	public Iterable<IAgent> allAgentsIterable() {
		return new Iterable<IAgent>() {

			@Override
			public Iterator<IAgent> iterator() {
				return agents.iterator();
			}
		};
	}

	public Iterable<IAgent> unfinishedAgentsIterable() {
		return new Iterable<IAgent>() {

			@Override
			public Iterator<IAgent> iterator() {
				return new Iterator<IAgent>() {

					private int index = 0;

					@Override
					public boolean hasNext() {
						if (index >= agents.size())
							return false;
						while (true) {
							if (agents.get(index).getIsFinished()) {
								index++;
								if (index >= agents.size())
									return false;
								continue;
							}
							return agents.size() > index;
						}
					}

					@Override
					public IAgent next() {
						return agents.get(index++);
					}
				};
			}
		};
	}

	public IAgent globalBestAgent() {
		if (this.globalBestIndex < 0 || this.globalBestIndex >= this.agents.size())
			return null;
		return agents.get(this.globalBestIndex);
	}

	public void refreshPopulationState() {
		findBest();
		adjustAgentDirVectors();
		refreshPopulationVector();
		refreshPopulationCenter();
		// ------------------------------------------------------------------------------

		// refresh population targets!!!!!!!!!!!! TODO
		// ------------------------------------------------------------------------------

	}

	private void adjustAgentDirVectors() {
		for(IAgent agent:agents){
			if(agent.colided())
				agent.getCurrentDirectionVector().set(0, 0).set(1, 0);
			else
				if(agent.getCurrentDirectionVector().norm()>this.maxPopulationSpeed)
					agent.getCurrentDirectionVector().normalise().scalarMultiply(maxPopulationSpeed);
		}
	}

	private void refreshPopulationCenter() {
		try {
			IVector newCenter = new Vector(0, 0);
			for (IAgent agent : agents) 
				newCenter.add(agent.getCurrentPosition());
			newCenter.scalarMultiply(1. / agents.size());
			this.centerPosition = newCenter;
		} catch (IncompatibleOperandException e) {
			return;
		}
	}

	private void findBest() {
		double bestFitness = -1;
		int bestIndex = 0;
		for (int i = 0; i < agents.size(); i++)
			if (bestFitness < agents.get(i).getFitness()) {
				bestFitness = agents.get(i).getFitness();
				bestIndex = i;
			}
		this.globalBestIndex = bestIndex;

		for (int agInd = 0; agInd < agents.size(); agInd++) {
			if (agInd == this.globalBestIndex)
				continue;
			else {
				agents.get(agInd).setPopBestAgent(agents.get(this.globalBestIndex));
			}
		}
	}

	private void refreshPopulationVector() {
		IVector popVector = new Vector(0, 0);
		for (IAgent agent : agents)
			try {
				popVector.add(agent.getCurrentDirectionVector());
			} catch (IncompatibleOperandException e) {
			}
		if (popVector.norm() > this.maxPopulationSpeed)
			popVector.normalise().scalarMultiply(this.maxPopulationSpeed);
		this.popVector = popVector;
		
	}

	public IVector getPopVector() {
		return this.popVector;
	}

	@Override
	public int numberOfAgents() {
		return this.agents.size();
	}

	@Override
	public IVector getCenterPosition() {
		return centerPosition;
	}

}
