package hr.fer.bachelorThesis.algorithm.strategy;

import java.util.LinkedList;
import java.util.List;

import hr.fer.bachelorThesis.algorithm.memory.ISharedMemory;
import hr.fer.bachelorThesis.algorithm.memory.MemoryElement;
import hr.fer.bachelorThesis.algorithm.pso.IAgent;
import hr.fer.bachelorThesis.algorithm.util.ELabyrinthDirection;
import hr.fer.bachelorThesis.algorithm.util.ParticleDirection;
import hr.fer.bachelorThesis.labyrinth.ELabyrinthElement;
import hr.fer.bachelorThesis.labyrinth.ILabyrinth;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.IncompatibleOperandException;
import hr.fer.bachelorThesis.util.linearAlgebra.Vector;
import hr.fer.bachelorThesis.util.random.IRNG;
import hr.fer.bachelorThesis.util.random.RNG;

public class FavourUnvisitedStrategy implements IAgentStrategy {

	private IAgentStrategy strategy;
	private ILabyrinth labyrinth;
	private ISharedMemory memory;
	private IRNG rand;
	private double unvisitedInfluence;

	public FavourUnvisitedStrategy(ILabyrinth labyrinth, ISharedMemory memory, double unvisitedInfluence,
			IAgentStrategy strategy) {
		this.strategy = strategy;
		this.labyrinth = labyrinth;
		this.memory = memory;
		this.unvisitedInfluence = unvisitedInfluence;
		this.rand = RNG.getRNG();
	}

	@Override
	public IVector nextDirectionVector(IAgent agent) throws IncompatibleOperandException {
		IVector thisStrategyVector = getDirVector(agent);
		//System.out.println("Unvisited  dir vector: "+thisStrategyVector);
		return strategy.nextDirectionVector(agent).add(thisStrategyVector);
	}


	private IVector getDirVector(IAgent agent) {
		List<IVector> neightList = getLegalNeighbourPositions(agent.getCurrentPosition());
		List<Double> fitList = calculatePositionsFitnessValues(neightList);
		
		sortDescneding(neightList, fitList);
		normaliseFitness(fitList); 
		
		IVector nextPosition = rouletteWheleDirection(neightList, fitList);
		IVector nextDirection = null;
		try {
			nextDirection = nextPosition.sub(agent.getCurrentPosition());
		} catch (IncompatibleOperandException ignorable) {
			// this code won't run since the vector dimensions are 
			// consistent throughout the algorithm
		}
		nextDirection.scalarMultiply(this.unvisitedInfluence);
		return nextDirection;
	}

	private IVector rouletteWheleDirection(List<IVector> neightList, List<Double> fitList) {
		double probability = rand.nextDouble();
		double lowerProbBound = 0;
		double upperProbBound = 0;
		for (int i = 0; i < fitList.size(); i++) {
			lowerProbBound = upperProbBound;
			upperProbBound += fitList.get(i);
			if (probability >= lowerProbBound && probability < upperProbBound)
				return neightList.get(i);
		}
		return neightList.get(0);
	}

	private void normaliseFitness(List<Double> fitList) {

		double avg = 0;
		for (Double f : fitList) {
			avg += f;
		}
		for (int i = 0; i < fitList.size(); i++) {
			fitList.set(i, fitList.get(i) / avg);
		}

	}

	private void sortDescneding(List<IVector> neightList, List<Double> fitList) {
		int size = fitList.size();
		IVector tmpV = null;
		double tmpD = 0;
		for (int outer = 0; outer < size - 1; outer++) {
			for (int inner = outer+1; inner < size; inner++) {

				if (fitList.get(outer) <= fitList.get(inner)) {
					// switch positions
					tmpV = neightList.get(outer);
					neightList.set(outer, neightList.get(inner));
					neightList.set(inner, tmpV);
					// switch fitness
					tmpD = fitList.get(outer);
					fitList.set(outer, fitList.get(inner));
					fitList.set(inner, tmpD);
				}

			}
		}
	}

	private List<Double> calculatePositionsFitnessValues(List<IVector> localPositions) {
		LinkedList<Double> fintessValues = new LinkedList<>();
		for (IVector p : localPositions) {
			MemoryElement memEle = this.memory.atPosition(p);
			fintessValues.add(memEle.calculatePositionScore());
		}
		return fintessValues;
	}

	private List<IVector> getLegalNeighbourPositions(IVector currentPosition) {
		LinkedList<IVector> directions = new LinkedList<>();
		for (ELabyrinthDirection eDir : ELabyrinthDirection.values()) {
			if(eDir==ELabyrinthDirection.BASE) continue;
			ParticleDirection dir = new ParticleDirection(eDir);

			IVector position = null;
			try {
				position = currentPosition.nAdd(new Vector(dir.toXOffset(), dir.toYOffset()));
			} catch (IncompatibleOperandException e) {
			}

			ELabyrinthElement labEle = labyrinth.valueAt(position);
			if(labEle == ELabyrinthElement.EXIT){
				directions.clear();
				directions.add(position);
				return directions;
			}
			if (labEle != ELabyrinthElement.WALL)
				directions.add(position);
		}

		return directions;
	}

}
