package hr.fer.bachelorThesis.labyrinth.generator;

import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Set;

import hr.fer.bachelorThesis.labyrinth.ELabyrinthElement;
import hr.fer.bachelorThesis.labyrinth.LabyrinthElement;
import hr.fer.bachelorThesis.util.linearAlgebra.IVector;
import hr.fer.bachelorThesis.util.linearAlgebra.Vector;

public class FixedEntranceLabyrinthGenerator implements ILabyrinthGeneratorStrategy {

	private class Point implements Comparable<Point> {
		int x;
		int y;
		int depth;

		public Point(int x, int y) {
			this.x = x;
			this.y = y;
			depth = 0;
		}

		@Override
		public String toString() {
			return x + " " + y;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + x;
			result = prime * result + y;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Point other = (Point) obj;
			if (x != other.x)
				return false;
			if (y != other.y)
				return false;
			return true;
		}

		@Override
		public int compareTo(Point o) {
			return Double.compare(depth, o.depth);
		}
	}

	private EEntranceLabel label;
	private int numberOfExits;

	public FixedEntranceLabyrinthGenerator(EEntranceLabel label, int numberOfExits) {
		this.label = label;
		this.numberOfExits = numberOfExits;
	}

	public List<IVector> generateLabyrinth(LabyrinthElement[][] labyrinthMatrix, int rows, int cols) {
		if (rows < 3 && cols < 3)
			throw new IllegalArgumentException("Labyrinth dimension is too small.");
		Random rand = new Random();

		Point[][] visited = new Point[rows][cols];
		Point startPoint = getStartPoint(label, rows, cols);

		layoutGenerator(labyrinthMatrix, rows, cols, label, rand, visited, startPoint);

		LinkedList<IVector> output = new LinkedList<>();
		
		addStartAndExit(labyrinthMatrix, rows, cols, numberOfExits, visited, startPoint, output);

		return output;
	}

	private void addStartAndExit(LabyrinthElement[][] labyrinthMatrix, int rows, int cols, int numberOfExits,
			Point[][] visited, Point startPoint, LinkedList<IVector> output) {

		labyrinthMatrix[startPoint.x][startPoint.y].set(ELabyrinthElement.START);
		output.addFirst(new Vector(startPoint.x, startPoint.y));

		addExits(labyrinthMatrix, rows, cols, visited, numberOfExits, output);

	}

	private void layoutGenerator(LabyrinthElement[][] labyrinthMatrix, int rows, int cols, EEntranceLabel label,
			Random rand, Point[][] visited, Point startPoint) {
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++)
				visited[i][j] = new Point(i, j);
		}

		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				labyrinthMatrix[i][j].set(ELabyrinthElement.PASSABLE);
			}
		}

		int mulRows = 3;
		int mulCols = 3;

		// vertical stripes: to rows-1 because the last was marked by frame
		for (int i = mulRows; i < rows - 1; i += mulRows) {
			for (int j = 0; j < cols; j++) {
				labyrinthMatrix[i][j].set(ELabyrinthElement.WALL);
			}
		}

		// horizontal stripes: to cols-1 becaus the last was marked by frame
		for (int j = mulCols; j < cols - 1; j += mulCols) {
			for (int i = 0; i < rows; i++) {
				labyrinthMatrix[i][j].set(ELabyrinthElement.WALL);
			}
		}

		Point firstPos = new Point((startPoint.x / mulRows) * mulRows + 1, (startPoint.y / mulCols) * mulCols + 1);

		recursiveLabyrinth(firstPos, mulRows, mulCols, labyrinthMatrix, visited, rows, cols, rand, 1);

		addFrame(labyrinthMatrix, rows, cols);
		return;
	}

	private void addExits(LabyrinthElement[][] labyrinthMatrix, int rows, int cols, Point[][] visited,
			int numberOfExits, LinkedList<IVector> output) {

		List<Point> listOfPoints = generateListOfPoints(visited, rows, cols);

		int loopDuration = (numberOfExits < listOfPoints.size()) ? numberOfExits : listOfPoints.size();

		for (int i = 0; i < loopDuration; i++) {
			// TODO add the defense against too many exits added!
			addExitToLocation(listOfPoints.get(i), labyrinthMatrix, rows, cols);

		}

	}

	private List<Point> generateListOfPoints(Point[][] visited, int rows, int cols) {
		Point[] arrayOfPoints = matrixAsList(visited, rows, cols);
		List<Point> listOfPoints = Arrays.asList(arrayOfPoints);
		listOfPoints.sort(new Comparator<Point>() {
			@Override
			public int compare(Point o1, Point o2) {
				return -Double.compare(o1.depth, o2.depth);
			}
		});

		int toIndex = 0;
		for (; toIndex < listOfPoints.size(); toIndex++)
			if (listOfPoints.get(toIndex).depth == 0)
				break;

		return listOfPoints.subList(0, toIndex);
	}

	private boolean addExitToLocation(Point point, LabyrinthElement[][] labyrinthMatrix, int rows, int cols) {
		Set<Point> closed;
		LinkedList<Point> open;
		if (labyrinthMatrix[point.x][point.y].equals(ELabyrinthElement.PASSABLE) && point.x != 0 && point.y != 0
				&& point.x != rows - 1 && point.y != cols - 1) {

			labyrinthMatrix[point.x][point.y].set(ELabyrinthElement.EXIT);
			return true;
		} else {
			closed = new HashSet<>();
			open = new LinkedList<>();
			open.addAll(expand(point, rows, cols));
		}

		while (true) {
			if (open.isEmpty())
				return false;

			Point p = open.removeFirst();
			if (closed.contains(p))
				continue;
			closed.add(p);

			if (labyrinthMatrix[p.x][p.y].equals(ELabyrinthElement.PASSABLE)) {
				labyrinthMatrix[p.x][p.y].set(ELabyrinthElement.EXIT);
				return true;
			}
			LinkedList<Point> succs = expand(p, rows, cols);
			open.addAll(succs);
			open.sort(new Comparator<Point>() {

				@Override
				public int compare(Point o1, Point o2) {
					double distanceThisToGoal = (o1.x - point.x) * (o1.x - point.x)
							+ (o1.y - point.y) * (o1.y - point.y);
					double distanceOtherToGoal = (o2.x - point.x) * (o2.x - point.x)
							+ (o2.y - point.y) * (o2.y - point.y);
					return Double.compare(Math.sqrt(distanceThisToGoal), Math.sqrt(distanceOtherToGoal));
				}
			});
		}

	}

	private LinkedList<Point> expand(Point p, int rows, int cols) {
		LinkedList<Point> succ = new LinkedList<>();
		for (int i = -1; i <= 1; i++) {
			for (int j = -1; j < 1; j++) {
				if (i == 0 && j == 0)
					continue;
				// we don't allow exits on board so >0 not >=0 and so on
				if (p.x + i > 0 && p.x + i < rows - 1 && p.y + j > 0 && p.y + j < cols - 1)
					succ.add(new Point(p.x + i, p.y + j));
			}
		}
		return succ;
	}

	private Point[] matrixAsList(Point[][] visited, int rows, int cols) {
		Point[] list = new Point[rows * cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++)
				list[i * rows + j] = visited[i][j];
		return list;
	}

	private Point getStartPoint(EEntranceLabel label, int rows, int cols) {
		switch (label) {
		case TOP_LEFT:
			return new Point(1, 1);
		case TOP_RIGHT:
			return new Point(1, cols - 2);
		case CENTER:
			return new Point((int) (rows / 2), (int) (cols / 2));
		case BOTTOM_LEFT:
			return new Point(rows - 2, 1);
		case BOTTOM_RIGHT:
			return new Point(rows - 2, cols - 2);
		default:
			return null;
		}
	}

	/**
	 * Recursive method for generating labyrinth. It uses depth first search
	 * implemented with recursion.Mazes generated with a depth-first search have
	 * a low branching factor and contain many long corridors, because the
	 * algorithm explores as far as possible along each branch before
	 * backtracking.
	 * 
	 * @param exitCandidates
	 */
	private boolean recursiveLabyrinth(Point point, int mulRows, int mulCols, LabyrinthElement[][] labyrinthMatrix,
			Point[][] visited, int rows, int cols, Random rand, int recursionDepth) {
		int x, y, a;
		List<Point> neigh = new LinkedList<>();

		// visited[point.x][point.y].depth = recursionDepth;
		visited[point.x][point.y].depth = 1; // must be a value bigger than zero

		getNeighbours(mulRows, mulCols, rows, cols, point, neigh);

		boolean succsCleared = false;

		while (neigh.size() > 0) {
			int randElem = rand.nextInt(neigh.size());
			if (visited[neigh.get(randElem).x][neigh.get(randElem).y].depth == 0) {
				x = neigh.get(randElem).x;
				y = neigh.get(randElem).y;

				if (x == point.x) { // same row
					for (int i = x; i <= x + mulRows - 2 && i < rows - 1; ++i) {
						a = (y + point.y) / 2 + mulRows / 2 - (mulRows + 1) % 2;
						labyrinthMatrix[i][a].set(ELabyrinthElement.PASSABLE);
						succsCleared = true;
					}
				} else { // same column
					for (int j = y; j <= y + mulCols - 2 && j < cols - 1; ++j) {
						a = (x + point.x) / 2 + mulCols / 2 - (mulCols + 1) % 2;
						labyrinthMatrix[a][j].set(ELabyrinthElement.PASSABLE);
						succsCleared = true;
					}
				}

				recursiveLabyrinth(new Point(x, y), mulRows, mulCols, labyrinthMatrix, visited, rows, cols, rand,
						recursionDepth + 1);

			}
			neigh.remove(randElem);
		}

		if (!succsCleared)
			visited[point.x][point.y].depth = recursionDepth; // must be a value
																// bigger than
																// zero

		return succsCleared;
	}

	/**
	 * 
	 * Returns reachable neighbours for passed `point'. There can be up to 4
	 * neighbours on north, east, west and south directions.
	 * 
	 * @param mul
	 *            Maze complexity
	 * @param point
	 *            Target point for whicht method calculates neighbours
	 * @param n
	 *            Vector to hold neighbours for current point
	 */
	private void getNeighbours(int mulRows, int mulCols, int rows, int cols, Point point, List<Point> n) {
		int x, y;
		if ((point.x - mulRows) > 0) {
			x = point.x - mulRows;
			y = point.y;
			n.add(new Point(x, y));
		}
		if ((point.x + mulRows) < rows) {
			x = point.x + mulRows;
			y = point.y;
			n.add(new Point(x, y));
		}
		if ((point.y - mulCols) > 0) {
			x = point.x;
			y = point.y - mulCols;
			n.add(new Point(x, y));
		}
		if ((point.y + mulCols) < cols) {
			x = point.x;
			y = point.y + mulCols;
			n.add(new Point(x, y));
		}
	}

	private void addFrame(LabyrinthElement[][] labyrinthMatrix, int rows, int cols) {
		// top & bottom//
		for (int i = 0; i < cols; ++i) {
			labyrinthMatrix[0][i].set(ELabyrinthElement.WALL);
			labyrinthMatrix[rows - 1][i].set(ELabyrinthElement.WALL);
		}

		// left & right//
		for (int i = 0; i < rows; ++i) {
			labyrinthMatrix[i][0].set(ELabyrinthElement.WALL);
			labyrinthMatrix[i][cols - 1].set(ELabyrinthElement.WALL);

		}
	}

}
